---------------------------------------------------------------------
-- File:    CRC_Check.vhd
-- Description:   CRC combinatorial calculation in one clock cycle
-- Designer:      Christos Zamantzas
-- Date:          24 July 2008 rev.11


-- modification 1.12.2005 : new frame configuration 
-- CID STAT1 STAT2 ...DATAS.... DAC12 DAC34 DAC56 DAC78 FID CRC1 CRC2

-- modifications: 01.02.2006 : cleaning

-- modifications: 02.02.2006 : ADDITIONS
--   Card ID (CID)    check    (frame 1)
--   Status           check    (frame 2-3)
--   Frame ID (FID)   check    (frame 18)

-- corrections: 06.02.2006 
--   CID, FID, Status included variables.

-- modifications: 23.07.2008 : 
--   change on how the SOF, DataValid are decoded
--   added differentiation on CRC and disconnection errors

-----------------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;

-- entity declaration

ENTITY CRC_Check IS
PORT
(
CLK         : IN STD_LOGIC;
status      : IN STD_LOGIC_VECTOR(1 downto 0);     --Status: 0x not valid D, 1x valid D, x0 not SOF, x1 SOF
RESET       : IN STD_LOGIC;
D           : IN STD_LOGIC_VECTOR(15 downto 0);
calcCRC     : OUT STD_LOGIC_VECTOR(31 downto 0);

-- Control & Status Outputs.
CRCout      : OUT STD_LOGIC_VECTOR(31 downto 0);
D16out      : OUT STD_LOGIC_VECTOR(15 downto 0); 
count       : OUT STD_LOGIC_VECTOR(4 downto 0); 
DValid      : OUT STD_LOGIC;                       -- =1 when not in IDLE state 
CRCready    : OUT STD_LOGIC;                       -- =1 when CRCout ready for compare
ERR         : OUT STD_LOGIC;                       -- =1 when different CRC was expected
ERR_TLK     : OUT STD_LOGIC;                       -- =1 when connection error 
CFCstatusA  : OUT STD_LOGIC_VECTOR(15 downto 0);   -- =1 when error detected in the tunnel(HT,PS,or CFC)
CFCstatusB  : OUT STD_LOGIC_VECTOR(15 downto 0);   -- =1 when error detected in the tunnel(HT,PS,or CFC)
FID         : OUT STD_LOGIC_VECTOR(15 downto 0);   -- Frame ID
CID         : OUT STD_LOGIC_VECTOR(15 downto 0);   -- Card ID
SOF         : OUT STD_LOGIC;                       -- =1 Indicates start of frame  
DAC_1_2x    : OUT STD_LOGIC_VECTOR(15 downto 0);
DAC_3_4x    : OUT STD_LOGIC_VECTOR(15 downto 0);
DAC_5_6x    : OUT STD_LOGIC_VECTOR(15 downto 0);
DAC_7_8x    : OUT STD_LOGIC_VECTOR(15 downto 0)
);

END CRC_Check;

-- architecture declaration

ARCHITECTURE CRC_CHECK_architecture OF CRC_Check IS

-- function declaration

-------------------------------------------------------------------------
-- credits for the nextCRC function goes to: 
-----------------------------------------------------------------------

-- File:  PCK_CRC32_D16.vhd              
-- Date:  Tue Sep 30 17:47:52 2003      
--     
-- Copyright (C) 1999-2003 Easics NV. 
-- This source file may be used and distributed without restriction    
-- provided that this copyright statement is not removed from the file 
-- and that any derivative work contains the original copyright notice
-- and the associated disclaimer.
--
-- THIS SOURCE FILE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS
-- OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
-- WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--
-- Purpose: VHDL package containing a synthesizable CRC function
--   * polynomial: (0 1 2 4 5 7 8 10 11 12 16 22 23 26 32)
--   * data width: 16
--     
-- Info: janz@easics.be (Jan Zegers)           
--       http://www.easics.com  
-----------------------------------------------------------------------

function nextCRC 
    ( Data:  std_logic_vector(15 downto 0);
      CRC:   std_logic_vector(31 downto 0) )
    return std_logic_vector is

    variable D: std_logic_vector(15 downto 0);
    variable C: std_logic_vector(31 downto 0);
    variable NewCRC: std_logic_vector(31 downto 0);

  begin

    D := Data;
    C := CRC;
      
    NewCRC(0) := D(12) xor D(10) xor D(9) xor D(6) xor D(0) xor C(16) xor C(22) xor C(25) xor C(26) xor C(28);
    NewCRC(1) := D(13) xor D(12) xor D(11) xor D(9) xor D(7) xor D(6) xor D(1) xor D(0) xor C(16) xor C(17) xor C(22) xor C(23) xor C(25) xor C(27) xor C(28) xor C(29);
    NewCRC(2) := D(14) xor D(13) xor D(9) xor D(8) xor D(7) xor D(6) xor D(2) xor D(1) xor D(0) xor C(16) xor C(17) xor C(18) xor C(22) xor C(23) xor C(24) xor C(25) xor C(29) xor C(30);
    NewCRC(3) := D(15) xor D(14) xor D(10) xor D(9) xor D(8) xor D(7) xor D(3) xor D(2) xor D(1) xor C(17) xor C(18) xor C(19) xor C(23) xor C(24) xor C(25) xor C(26) xor C(30) xor C(31);
    NewCRC(4) := D(15) xor D(12) xor D(11) xor D(8) xor D(6) xor D(4) xor D(3) xor D(2) xor D(0) xor C(16) xor C(18) xor C(19) xor C(20) xor C(22) xor C(24) xor C(27) xor C(28) xor C(31);
    NewCRC(5) := D(13) xor D(10) xor D(7) xor D(6) xor D(5) xor D(4) xor D(3) xor D(1) xor D(0) xor C(16) xor C(17) xor C(19) xor C(20) xor C(21) xor C(22) xor C(23) xor C(26) xor C(29);
    NewCRC(6) := D(14) xor D(11) xor D(8) xor D(7) xor D(6) xor D(5) xor D(4) xor D(2) xor D(1) xor C(17) xor C(18) xor C(20) xor C(21) xor C(22) xor C(23) xor C(24) xor C(27) xor C(30);
    NewCRC(7) := D(15) xor D(10) xor D(8) xor D(7) xor D(5) xor D(3) xor D(2) xor D(0) xor C(16) xor C(18) xor C(19) xor C(21) xor C(23) xor C(24) xor C(26) xor C(31);
    NewCRC(8) := D(12) xor D(11) xor D(10) xor D(8) xor D(4) xor D(3) xor D(1) xor D(0) xor C(16) xor C(17) xor C(19) xor C(20) xor C(24) xor C(26) xor C(27) xor C(28);
    NewCRC(9) := D(13) xor D(12) xor D(11) xor D(9) xor D(5) xor D(4) xor D(2) xor D(1) xor C(17) xor C(18) xor C(20) xor C(21) xor C(25) xor C(27) xor C(28) xor C(29);
    NewCRC(10) := D(14) xor D(13) xor D(9) xor D(5) xor D(3) xor D(2) xor  D(0) xor C(16) xor C(18) xor C(19) xor C(21) xor C(25) xor C(29) xor C(30);
    NewCRC(11) := D(15) xor D(14) xor D(12) xor D(9) xor D(4) xor D(3) xor D(1) xor D(0) xor C(16) xor C(17) xor C(19) xor C(20) xor C(25) xor C(28) xor C(30) xor C(31);
    NewCRC(12) := D(15) xor D(13) xor D(12) xor D(9) xor D(6) xor D(5) xor D(4) xor D(2) xor D(1) xor D(0) xor C(16) xor C(17) xor C(18) xor C(20) xor C(21) xor C(22) xor C(25) xor C(28) xor C(29) xor C(31);
    NewCRC(13) := D(14) xor D(13) xor D(10) xor D(7) xor D(6) xor D(5) xor D(3) xor D(2) xor D(1) xor C(17) xor C(18) xor C(19) xor C(21) xor C(22) xor C(23) xor C(26) xor C(29) xor C(30);
    NewCRC(14) := D(15) xor D(14) xor D(11) xor D(8) xor D(7) xor D(6) xor D(4) xor D(3) xor D(2) xor C(18) xor C(19) xor C(20) xor C(22) xor C(23) xor C(24) xor C(27) xor C(30) xor C(31);
    NewCRC(15) := D(15) xor D(12) xor D(9) xor D(8) xor D(7) xor D(5) xor D(4) xor D(3) xor C(19) xor C(20) xor C(21) xor C(23) xor C(24) xor C(25) xor C(28) xor C(31);
    NewCRC(16) := D(13) xor D(12) xor D(8) xor D(5) xor D(4) xor D(0) xor C(0) xor C(16) xor C(20) xor C(21) xor C(24) xor C(28) xor C(29);
    NewCRC(17) := D(14) xor D(13) xor D(9) xor D(6) xor D(5) xor D(1) xor C(1) xor C(17) xor C(21) xor C(22) xor C(25) xor C(29) xor C(30);
    NewCRC(18) := D(15) xor D(14) xor D(10) xor D(7) xor D(6) xor D(2) xor C(2) xor C(18) xor C(22) xor C(23) xor C(26) xor C(30) xor C(31);
    NewCRC(19) := D(15) xor D(11) xor D(8) xor D(7) xor D(3) xor C(3) xor C(19) xor C(23) xor C(24) xor C(27) xor C(31);
    NewCRC(20) := D(12) xor D(9) xor D(8) xor D(4) xor C(4) xor C(20) xor C(24) xor C(25) xor C(28);
    NewCRC(21) := D(13) xor D(10) xor D(9) xor D(5) xor C(5) xor C(21) xor C(25) xor C(26) xor C(29);
    NewCRC(22) := D(14) xor D(12) xor D(11) xor D(9) xor D(0) xor C(6) xor C(16) xor C(25) xor C(27) xor C(28) xor C(30);
    NewCRC(23) := D(15) xor D(13) xor D(9) xor D(6) xor D(1) xor D(0) xor C(7) xor C(16) xor C(17) xor C(22) xor C(25) xor C(29) xor C(31);
    NewCRC(24) := D(14) xor D(10) xor D(7) xor D(2) xor D(1) xor C(8) xor C(17) xor C(18) xor C(23) xor C(26) xor C(30);
    NewCRC(25) := D(15) xor D(11) xor D(8) xor D(3) xor D(2) xor C(9) xor C(18) xor C(19) xor C(24) xor C(27) xor C(31);
    NewCRC(26) := D(10) xor D(6) xor D(4) xor D(3) xor D(0) xor C(10) xor C(16) xor C(19) xor C(20) xor C(22) xor C(26);
    NewCRC(27) := D(11) xor D(7) xor D(5) xor D(4) xor D(1) xor C(11) xor C(17) xor C(20) xor C(21) xor C(23) xor C(27);
    NewCRC(28) := D(12) xor D(8) xor D(6) xor D(5) xor D(2) xor C(12) xor C(18) xor C(21) xor C(22) xor C(24) xor C(28);
    NewCRC(29) := D(13) xor D(9) xor D(7) xor D(6) xor D(3) xor C(13) xor C(19) xor C(22) xor C(23) xor C(25) xor C(29);
    NewCRC(30) := D(14) xor D(10) xor D(8) xor D(7) xor D(4) xor C(14) xor C(20) xor C(23) xor C(24) xor C(26) xor C(30);
    NewCRC(31) := D(15) xor D(11) xor D(9) xor D(8) xor D(5) xor C(15) xor C(21) xor C(24) xor C(25) xor C(27) xor C(31);
    return NewCRC;

  end nextCRC;


-------------------------------------------------------------------------
-- Main Program

signal   Cnt      :integer range 0 to 31;
signal   D16_int  :std_logic_vector(15 downto 0);

begin

CRC_parallel : process( CLK, RESET )

variable D16      :std_logic_vector(15 downto 0);
variable oldD16   :std_logic_vector(15 downto 0);
variable CRC32    :std_logic_vector(31 downto 0);
variable CRCrx    :std_logic_vector(31 downto 0);
variable S        :std_logic_vector(1 downto 0);
variable WordCnt  :integer range 0 to 21;

begin
   if RESET='1' 
      then
         CRC32    := (others => '1');
         WordCnt  := 0;
         DValid   <= '0';
         CRCready <= '0';
         ERR      <= '0';
         ERR_TLK  <= '0';
         SOF      <= '0';
         
      elsif CLK'event and CLK='1' 
         then
            S := status(1 downto 0);
            D16 := D(15 downto 0);
            case S is
               when "00" =>
                  WordCnt := 0;
                  DValid  <= '0';                     -- IDLE
                  SOF     <= '0';
                  ERR_TLK <= '0';
                  
               when "01" => 
                  CRC32 := (others => '1'); 
                  if WordCnt = 0 then
                     WordCnt  := 1;
                     DValid   <= '0';                 -- SOF (NOT IDLE)
                     SOF      <= '1';
                     CRCready <= '0';
                     ERR      <= '0';
                  else 
                     WordCnt  := 21;
                     DValid   <= '0';                 -- ERROR (SOF has been received in the middle of data)
                     SOF      <= '0';
                     CRCready <= '0';
                     ERR_TLK  <= '1';   
                  end if;
                  
               when "10" => 
                  DValid <= '0';                      -- DATA (NOT IDLE)
                  SOF    <= '0';
                  if WordCnt > 0 then
                     if WordCnt < 20 then  
                        DValid <= '1'; 
                     end if;
                     case WordCnt is 
                        when 19 =>                    -- CRC's 1st part  
                           oldD16 := D16;
                        when 20 =>                    -- CRC's 2nd part  
                           CRCrx := oldD16 & D16;
                           calcCRC <= CRCrx;          -- out for comparison
                           CRCready <= '1';
                           
                           ERR <= '0';                -- check CRC end 
                           if (CRC32 /= CRCrx) then 
                              ERR <= '1';
                           end if;
                        when others => 
                           CRC32 := nextCRC(D16,CRC32); 
                     end case;
                     WordCnt := WordCnt + 1;
                  end if;
                 
               when others =>                         -- ERROR 
                  CRC32    := (others => '1');
                  WordCnt  := 21;                     -- mod.20080723 (was: =0)
                  DValid   <= '0'; 
                  SOF      <= '0';
                  CRCready <= '0';            
                  ERR_TLK  <= '1';
            end case;
      end if;  
      CRCout  <= CRC32;
      D16out  <= D16;
      D16_int <= D16;
      Cnt     <= WordCnt - 1;
   
end process CRC_parallel;

process (clk)
begin
   if CLK'event and CLK='1' then
      Count <= std_logic_vector(to_signed((Cnt),5));
      case Cnt is 
         when  1 => CID         <= D16_int;   
         when  2 => CFCstatusA  <= D16_int;
         when  3 => CFCstatusB  <= D16_int;
         when 14 => DAC_1_2x    <= D16_int;
         when 15 => DAC_3_4x    <= D16_int;
         when 16 => DAC_5_6x    <= D16_int;
         when 17 => DAC_7_8x    <= D16_int;
         when 18 => FID         <= D16_int;
         when others =>
      end case; 
   end if;
end process;

end CRC_CHECK_architecture;