--------------------------------------------------------
-- FILE: Range_ADC.vhd
-- Author: Christos Zamantzas
--------------------------------------------------------
-- Description:
--   The ADC values arriving correspont to measurements of the discharge level of the capacitor of the CFC circuit.
--   Since the charge - discharge cycle does not span accross the whole range, this block:
--     - Calculates and keeps the min and max values seen over a period of time 
--     - Partially corrects by scalling down the value, i.e subtracting from it the minimum seen.
--     - Provides the ADC range (i.e. max-min seen) to the DataCombine block.
--------------------------------------------------------


-- Version: 2.4
-- DATE: 06/07/2009

-- ver. 2.1 - First version in VHDL
-- ver. 2.2 - Major change of logic: 
               -- Integrated here the ScaleDown.vhd block
               -- Removed adc_subtract, adc_max, adc_min components
               -- Added limits in the MIN & MAX values
              
-- ver. 2.3 - Major change of logic:
               -- re-arranged input/output signals
               -- renamed input/output signals
               -- removed one delay from the read enable output
               -- added numeric and std_unsigned libraries
               
-- ver. 2.4 - Changes for readability
                              
-- TODO: 
        -- remove timer LPM
        -- TBD:: add input to **** update/stop range calculation.

LIBRARY ieee;
USE ieee.std_logic_1164.all; 
USE IEEE.NUMERIC_STD.all;
USE IEEE.STD_LOGIC_UNSIGNED.all;

ENTITY Range_ADC IS 
   PORT
   (
    Data_in    :  IN  STD_LOGIC_VECTOR(19 DOWNTO 0); -- data input
    rden_in    :  IN  STD_LOGIC;                     -- new data available
    clk        :  IN  STD_LOGIC;

    Data_out   :  OUT STD_LOGIC_VECTOR(19 DOWNTO 0); -- data output
    rden_out   :  OUT STD_LOGIC;                     -- new data output avaialble
    ADC_range  :  OUT STD_LOGIC_VECTOR(11 DOWNTO 0)
   );
END Range_ADC;

ARCHITECTURE RTL OF Range_ADC IS 

COMPONENT range_calc_timer
   PORT(
       clock : IN STD_LOGIC;
       cout  : OUT STD_LOGIC;
       q     : OUT STD_LOGIC_VECTOR(31 DOWNTO 0));
END COMPONENT;

SIGNAL   Data_in_Counts :  STD_LOGIC_VECTOR(7 DOWNTO 0);
SIGNAL   Data_in_ADC    :  STD_LOGIC_VECTOR(11 DOWNTO 0);

SIGNAL   ADC_max        :  STD_LOGIC_VECTOR(11 DOWNTO 0);
SIGNAL   ADC_min        :  STD_LOGIC_VECTOR(11 DOWNTO 0);
SIGNAL   ADC_range_temp :  STD_LOGIC_VECTOR(11 DOWNTO 0);

SIGNAL   ADC_min_latch  :  STD_LOGIC_VECTOR(11 DOWNTO 0);
SIGNAL   update_range   :  STD_LOGIC;

SIGNAL   ADC_corr       :  STD_LOGIC_VECTOR(11 DOWNTO 0);

BEGIN 


-- counter enabled continously; cary out updates the range seen downstream. 
inst3 : range_calc_timer
PORT MAP(
        clock => clk,
        cout  => update_range
        );

PROCESS(clk)
BEGIN

IF (RISING_EDGE(clk)) THEN

   rden_out <= rden_in;
   
   IF (rden_in = '1') THEN
      -- Register the input
      Data_in_ADC    <= Data_in(11 DOWNTO 0);
      Data_in_Counts <= Data_in(19 DOWNTO 12);
      
      -- Scale Down the ADC input by the minimum ADC value calculated 
      ADC_corr <= Data_in(11 DOWNTO 0) - ADC_min_latch; 
--      IF Data_in(11 DOWNTO 0) < ADC_min_latch THEN
      IF Data_in < ADC_min_latch THEN
        ADC_corr <= (OTHERS =>'0');  
      END IF;
      
      -- calculate MAX ADC value seen
      IF (Data_in_ADC > ADC_max) THEN
         ADC_max <= Data_in_ADC;
      END IF;
      -- calculate MIN ADC value seen
      IF (Data_in_ADC < ADC_min) THEN
         ADC_min <= Data_in_ADC;
      END IF;  
      -- calculate the ADC range
      ADC_range_temp <= ADC_max - ADC_min;
      
   END IF;

   -- Regular update of the ADC range value for downstream usage
   IF (update_range = '1') THEN
    
      -- register (dffe) the minimum and the range seen for the ADC
      ADC_min_latch <= ADC_min;
      ADC_range <= ADC_range_temp;
    
      -- reset MIN & MAX values and restart calculation
      ADC_max <= X"F00"; --was: ADC_max <= "000000000000";
      ADC_min <= X"0FF"; --was: ADC_min <= "111111111111";
    
   END IF;

END IF;
END PROCESS;

Data_out <= (Data_in_Counts & ADC_corr);

END RTL;